#ifndef __SCHEDULE_H
#define __SCHEDULE_E

#include <__signedbv.h>
#include <iostream>
#include <fstream>
#include <sstream>
#include <stdlib.h>
#include <set>
#include <map>
#include <list>
#include <unistd.h>
#include <stdio.h>
#define DEBUG

#define IF_DEBUG(x) 

inline __signedbv<32> __scoot_next_schedule(bool (**palt_table)(), bool * pindep_table, __signedbv<32> n )
{

	static std::set<int> runnable;
	static std::set<int> sleeps;
	static std::ofstream pout;
	static std::ifstream pin;

	// initialization code
	static bool initialized = false;
	if(!initialized)
	{
#if 0
		unsigned seed;	
		std::cout << "Random seed? ";
		std::cin  >> seed;
		
		srand(seed);
#endif
		// all processes are runnable
		for(int i = 0; i < n.to_int(); i++)
		{
			runnable.insert(i);
		}

		pout.open("./scoot_sim.pipe.out", std::ofstream::out);
		pin.open("./scoot_sim.pipe.in", std::ifstream::in);
		initialized = true;
	}


	IF_DEBUG(std::cout << "Runnable processes: " << runnable.size() << std::endl);

	// end of evaluation phase?
	if (runnable.empty())
	{
		// all processes are runnable
		for(int i = 0; i < n.to_int(); i++)
		{
			runnable.insert(i);
		}

		sleeps.clear();

		return __signedbv<32>(-1);
	}


	// run independant processes first
	{
		std::set<int> independant = runnable;

		for(std::set<int>::const_iterator it1  = runnable.begin();
			it1 != runnable.end();
			it1++)
		{
			int p1 = *it1;
			for(std::set<int>::const_iterator it2  = runnable.begin();
			it2 != runnable.end();
			it2++)
			{
				int p2 = *it2;

				if(p1 == p2)
					continue;
				bool indep = pindep_table[p1*n.to_int()+p2]; 
				if(!indep)
				{
					independant.erase(p1);
					independant.erase(p2);
				}
			}
		}

		IF_DEBUG(std::cout << "Independant proc. : " << independant.size() << std::endl);

		if(!independant.empty())
		{
			int pid = *independant.begin();
			runnable.erase(pid);

			IF_DEBUG(td::cout << "scheduling " << pid << std::endl);
			return __signedbv<32>(pid);
		}
	}

	assert(!runnable.empty());

	// Persistent
	std::set<int> persistents;

#ifdef PERSISTENT_SET
	for(std::set<int>::const_iterator it1 = runnable.begin();
			it1 != runnable.end(); it1++)
	{
		int pi = *it1;
		for(std::set<int>::const_iterator it2 = runnable.begin();
			it2 != it1; it2++)
		{
			int pj = *it2;
			bool (*pfunc)() = palt_table[pi*n.to_int()+pj]; 
			bool indep = (*pfunc)();
			if(indep)
			{
				IF_DEBUG(std::cout <<"Indep " << pi  << ", " << pj << std::endl);
				if(persistents.count(pi) == 0)
					persistents.insert(pj);
			}
			else
			{
				persistents.insert(pi);
				persistents.insert(pj);
			}

		}
	}
#else
	persistents = runnable;
#endif

	IF_DEBUG(std::cout <<"Persistent set: " << persistents.size() << " proc." << std::endl);
	IF_DEBUG(std::cout <<"Sleep set: " << sleeps.size() << " proc." << std::endl);

	std::set<int> awakes;

	for(std::set<int>::iterator it = persistents.begin(); it != persistents.end(); it++)
	{
		if(sleeps.count(*it) == 0)
			awakes.insert(*it);
	}

	if(awakes.empty())
	{
		IF_DEBUG(std::cout << "Set awakes is empty" << std::endl);
		exit(0);
	}
	else if(awakes.size() == 1)
	{ // only one process to run?
		int pid = *runnable.begin();
		runnable.erase(pid);
		IF_DEBUG(std::cout << "scheduling " << pid << std::endl);
		return __signedbv<32>(pid);
	}

	std::map<int, std::set<int> > next_sleeps;

#ifdef SLEEP_SET
	for(std::set<int>::const_iterator it1 = awakes.begin();
			it1 != awakes.end(); it1++)
	{
		int pi = *it1;
		for(std::set<int>::const_iterator it2 = sleeps.begin();
			it2 != sleeps.end(); it2++)
		{
			int pj = *it2;
			bool (*pfunc)() = palt_table[pi*n.to_int()+pj]; 
			bool indep = (*pfunc)();
			if(indep) next_sleeps[pi].insert(pj);
		}
		sleeps.insert(pi);
	}
#endif

#ifdef RANDOM_SEARCH
	// pickup process randomly
	{
		int rnd = random();
		std::cout << "rnd " << rnd <<std::endl;
		int inc = RAND_MAX/awakes.size();
		std::set<int>::const_iterator pit = awakes.begin();
		for(int counter = inc; counter < rnd; counter += inc)
		{
			pit++;
		}
		assert( pit != awakes.end());
		sleeps = next_sleeps[*pit];
		runnable.erase(*pit);

		IF_DEBUG(std::cout << "scheduling " << *pit << std::endl);
		return __signedbv<32>(*pit);
	}
#else
/*		std::cout << "Simulator: Exploration choices (";
		for(std::set<int>::const_iterator ita = awakes.begin();
				ita != awakes.end(); ita++)
		{
			std::cout << *ita << " ";
		}
		std::cout << ") " << std::endl; */
		IF_DEBUG(std::cout << "Simulator sending " << awakes.size() << std::endl);
		pout << awakes.size() << std::endl;
		pout.flush();
		int next;
		IF_DEBUG(std::cout << "Simulator: Waiting for reply\n");
		pin >> next;
		IF_DEBUG(std::cout << "Simulator: recieved " << next << std::endl);

		assert(next >=0 && next < awakes.size());
		std::set<int>::iterator ita = awakes.begin();
		for(int i =0; i < next; i++)ita++;

		IF_DEBUG(std::cout << "scheduling " << *ita << std::endl);

		runnable.erase(*ita);
		sleeps = next_sleeps[*ita];
	

		return __signedbv<32>(*ita);
#endif

}

#endif
